%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%    This file describes the format of each of the instructions.           %
%    To simplify matters, each possible format is named with an ID.        %
%    That ID is assigned one template which can be used to print the       %
%    instruction, and another template which can be used to deposit the    %
%    instruction in memory.                                                %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%   The following formats are named as they are on the System/370
%     Reference Summary (Yellow card):
%
%   RR  :   R1,R2
%   RX  :   R1,D2(X2,B2)
%   RS  :   R1,R3,D2(B2)
%   SI  :   D1(B1),I2
%   S   :   D2(B2)
%
%   Both of these formats are called SS by The Reference Summary:
%
%   SS1 :   D1(L,B1),D2(B2)
%   SS2 :   D1(L1,B2),D2(L2,B2)
%
%   These formats are described but not named by the Reference Summary:
%
%   variant of  my name  explicit format      used by
%   ----------  -------  -------------------  -------
%        RR       RR1    R1                   (SPM)
%                 RR2    I1                   (SVC)
%                 MR1    M1,R2                (BC)
%                 MR2    R2                   (BC derivatives)
%
%        RX       MX1    M1,D2(X2,B2)         (BCR)
%                 MX2    D2(X2,B2)            (BCR derivatives)
%
%        RS       RSM    R1,M3,D2(B2)         (CLM,ICM,STCM)
%                 RS2    R1,D2(B2)            ..?..
%
%        SS       SS3    D1(L1,B2),D2(B2),I3  (SPK)
%
%        S        S0                          ..?..     (2 byte opcode)
%
%        S        S1     D1(B1)               ..?..     (1 byte opcode)
%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                          %
%  The AsmInstructionFormat is a list of the form:                         %
%                                                                          %
%    (<number of opcodes> . <field format>'s)                              %
%                                                                          %
%  If the number of received opcodes differs from the number specified,    %
%   an error message will be generated.                                    %
%                                                                          %
%  Otherwise the fluid variables Operand!*1, Operand!*2, etc. are bound    %
%   to the received operands.                                              %
%                                                                          %
%  Each field format is then evaluated and the result printed.             %
%                                                                          %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                          %
%  The primitive formatting functions are:                                 %
%                                                                          %
%  Name of function:    Used to prepare one of these formatted strings:    %
%  -------------------- -----------------------------------------------    %
%  AsmFormatRegister        <register name or number>                      %
%  AsmFormatImmediateByte   <8-bit immediate quantity>                     %
%  AsmFormatDispBase        <displacement>(<base register>)                %
%                           <displacement>                                 %
%  AsmFormatDispIndexBase   <displacement>(<index reg>,<base register>)    %
%                           <displacement>(<base register>)                %
%                           <displacement>(<index register>)               %
%                           <displacement>                                 %
%  AsmFormatFourBitMask     <4-bit mask>                                   %
%  AsmFormatDispLengthBase  <displacement>(<length>,<base register>)       %
%                           <displacement>(<length>)                       %
%                                                                          %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(deflist '((rr (2 (asmformatregister operand*1) "," 
                      (asmformatregister operand*2))) 

            (rr1 (1 (asmformatregister operand*1))) 

            (rr2 (1 (asmformatimmediatebyte operand*1))) 

            (mr1 (2 (asmformatfourbitmask operand*1) "," 
                      (asmformatregister operand*2))) 

            (mr2 (1 (asmformatregister operand*1))) 

            (rx (2 (asmformatregister operand*1) "," 
                      (asmformatdispindexbase operand*2))) 

            (mx1 (2 (asmformatfourbitmask operand*1) "," 
                      (asmformatdispindexbase operand*2))) 

            (mx2 (1 (asmformatdispindexbase operand*1))) 

            (rs (3 (asmformatregister operand*1) "," 
                      (asmformatregister operand*2) "," 
                      (asmformatdispbase operand*3))) 

            (rsm (3 (asmformatregister operand*1) "," 
                      (asmformatfourbitmask operand*2) "," 
                      (asmformatdispbase operand*3))) 

            (rs2 (2 (asmformatregister operand*1) "," 
                      (asmformatdispbase operand*2))) 

            (si (2 (asmformatdispbase operand*1) "," 
                      (asmformatimmediatebyte operand*2))) 

            (ss1 (2 (asmformatdisplengthbase operand*1 256) "," 
                      (asmformatdispbase operand*2))) 

            (ss2 (2 (asmformatdisplengthbase operand*1 16) "," 
                      (asmformatdisplengthbase operand*2 16))) 

            (ss3 (3 (asmformatdisplengthbase operand*1 16) "," 
                      (asmformatdispbase operand*2) "," 
                      (asmformatfourbitmask operand*3))) 

            (s (1 (asmformatdispbase operand*1))) 

            (s1 (1 (asmformatdispbase operand*1))) 

            (s0 (0)) 

            (diagnose (2                % this might not work....
                   (asmformatimmediatebyte operand*1) 
                      (asmformatdefineconstant operand*2))) 

            (entry (1 (asmformatentry operand*1))) 

            (using (2 (asmformatlabel operand*1) "," 
                      (asmformatregister operand*2))) 

            (ds (1 (asmformatdefinestorage operand*1))) 
            (dc (1 (asmformatdefineconstant operand*1))) 

            (problem (nil (asmformatproblem operands*)))) 
                      % nil => any number of args, Operands!* = all args

            
         'asminstructionformat)

(put 'entry 'opcodetype 'entry)
(put 'using 'opcodetype 'using)
(put 'ds 'opcodetype 'ds)
(put 'dc 'opcodetype 'dc)
(put 'problem 'opcodetype 'problem)


(de setcodes (n type opcodes) 
 (foreach opcode in opcodes do 
 (progn (put opcode 'opcodenumber n) 
     (put opcode 'opcodetype type) 
     (setq n (plus n 1)))))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                       RR format                                   (1-63) %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(setcodes 5 'rr '(balr bctr junk        %   0-  7
                    ssk isk junk junk junk junk mvcl clcl 	%   8- 15
                    lpr lnr ltr lcr nr clr or xr 	%  16- 23
                    lr cr ar sr mr dr alr slr 	%  24- 31
                    lpdr lndr ltdr lcdr hdr lrdr mxr mxdr 	%  32- 39
                    ldr cdr adr sdr mdr ddr awr swr 	%  40- 47
                    lper lner lter lcer her lrer axr sxr 	%  48- 55
                    ler cer aer ser mer der aur sur))	%  56- 63

(setcodes 4 'rr1 '(spm))
(setcodes 7 'mr1 '(bcr))
(setcodes 10 'rr2 '(svc))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                       RX format                                 (64-127) %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(setcodes 64 'rx '(sth la stc ic ex bal bct junk 	%  64- 71
                     lh ch ah sh mh junk cvd cvb 	%  72- 79
                     st junk junk junk n cl o x 	%  80- 87
                     l c a s m d al sl  %  88- 95
                     std junk junk junk junk junk junk mxd 	%  96-103
                     ld cd ad sd md dd aw sw 	% 104-111
                     ste junk junk junk junk junk junk junk 	% 112-119
                     le ce ae se me de au su))	% 120-127

(setcodes 7 'mx1 '(bc))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                       S, SI, RS, RX formats                    (128-191) %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(setcodes 128 's1 '(ssm junk lpsw))     % 128-
(setcodes 131 'diagnose '(diagnose))    %    :
(setcodes 132 'si '(wrd rdd))           %    :
(setcodes 134 'rs '(bxh bxle))          %    -135
(setcodes 136 'rs2 '(srl sll sra sla srdl sldl srda slda))	% 136-143
(setcodes 144 'rs '(stm))               % 144-
(setcodes 145 'si '(tm mvi))            %    :
(setcodes 147 's1 '(ts))                %    :
(setcodes 148 'si '(ni cli oi xi))      %    -151
(setcodes 152 'rs '(lm))                % 152-167
(setcodes 172 'si '(stnsm stosm))       % 168-
(setcodes 174 'rs '(sigp))              %    :
(setcodes 175 'si '(mc))                %    -175
(setcodes 177 'rx '(lra))               % 176-
(setcodes 182 'rs '(stctl lctl          %    -183
                      junk junk cs cds))% 184-
(setcodes 189 'rsm '(clm stcm icm))     %    -191

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                       S2 FORMATS                                         %
%  Some S format instructions have two byte opcodes.                       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 16#9C = 156

(setcodes 39936 's '(sio siof))         % 9C00-9C01
(setcodes 40192 's '(tio clrio))        % 9D00-9D01
(setcodes 40448 's '(hio hdv))          % 9E00-9E01
(setcodes 40704 's '(tch))              % 9F00

(setcodes 45570 's 
          '(stidp stidc sck stck sckc stckc 	% B202-B207
            spt stpt spka))             % B208-B20A

(setcodes 45579 's0 
          '(ipk junk ptlb))             % B20B-B20D

(setcodes 45584 's 
          '(spx stpx stap rrb))         % B210-B213
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                       SS formats                               (192-255) %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

                                                                  % 192-207
(setcodes 209 'ss1 '(mvn mvc mvz nc clc oc xc))	% 208-215
(setcodes 220 'ss1 '(tr trt ed edmk))   % 216-223
                                                                  % 220-239
(setcodes 240 'ss3 '(srp))              % 240-
(setcodes 241 'ss2 '(mvo pack unpk))    %    -247
(setcodes 248 'ss2 '(zap cp ap sp mp dp))	% 248-255


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                       BRANCH OPCODES                                     %
%                                                                          %
%  Special mnemonics for branch opcodes with implied mask fields.          %
%                                                                          %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de setbcopcodes (op1 op2 n) 
 (progn (put op1 'opcodenumber 7)       % get( 'BCR, 'OpCodeNumber )
     (put op2 'opcodenumber 71)         % get( 'BC,  'OpCodeNumber )
     (put op1 'opcodetype 'mr2) 
     (put op2 'opcodetype 'mx2) 
     (put op1 'mask n) 
     (put op2 'mask n) nil))
 

(setbcopcodes 'br 'b 15)                %  always
(setbcopcodes 'nopr 'nop 0)             %  never

(setbcopcodes 'bhr 'bh 2)               %  high
(setbcopcodes 'blr 'bl 4)               %  low
(setbcopcodes 'ber 'be 8)               %  equal
(setbcopcodes 'bnhr 'bnh 13)            %  not high
(setbcopcodes 'bnlr 'bnl 11)            %  not low
(setbcopcodes 'bner 'bne 7)             %  not equal
(setbcopcodes 'bor 'bo 1)               %             overflow    ones
(setbcopcodes 'bpr 'bp 2)               %             plus
(setbcopcodes 'bmr 'bm 4)               %             minus       mixed
(setbcopcodes 'bzr 'bz 8)               %             zero        zeros
(setbcopcodes 'bnpr 'bnp 13)            %             not plus
(setbcopcodes 'bnmr 'bnm 11)            %             not minus
(setbcopcodes 'bnzr 'bnz 7)             %             not zero
(setbcopcodes 'bnor 'bno 14)%                         not ones


