%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:BASE-IO.SL
% Title:        System independent IO routines for PSL
% Author:       Modified by Robert R. Kessler
% Created:      Modified 16 August 1982
% Modified:     20-Nov-84 13:17:17 (Brian Beach)
% Status:       Experimental
% Mode:         Lisp
% Package:      Kernel
% Compiletime:  PL:FAST-VECTOR.B PL:IF-SYSTEM.B PL:IO-DECLS.B
% Runtime:
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1982, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 20-Oct-86 W. Neun,ZIB Berlin
%  changed independentwritechar in favour of preloading                 more
% 20-Jun-86 W. Neun,ZIB Berlin
%  Added count feature for parens                                       more
% 20-Nov-84 13:14:53 (Brian Beach)
%  Changed INDEPENDENTWRITECHAR so that the buffer overflow is computed more
%  accurately.  The Number of characters in the buffer when the buffer is
%  written should now be exactly equal to the declared buffer size (instead
%  of two more, as it has been in the past.)
% 17-Jul-84 14:00 (Brian Beach)
%  Added missing STRINF's.
% 31-May-84 09:13:29 (Brian Beach)
%  Removed if-system for Chipmunk that handled tabs.  Exported the code
%  to the system specific terminal driver (in PHPP:HP.TEXT).
% 27-Mar-84 14:48:03 (Jim Ambras/CRC)
%  Added compiletime load of io-decls.
% 02-Dec-83 10:42:46 (Jim Ambras)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%==============================================================================
%
% Each individual system must have the following routines defined.
%
%   The following definitions are used in the routines:
%    FileDescriptor - A machine dependent word that references a file once
%      opened; generated by the Open
%    FileName - A Lisp string of the file name.
%
%  FileDescriptor := SysOpenRead (Channel,FileName);
%                                             % Open FileName for input and
%      % return a file descriptor used
%      % in later references to the
%      % file. Channel used only
%                                             % if needed to generate FileDesc
%  FileDescriptor := SysOpenWrite (Channel,FileName);
%                                             % Open FileName for output and
%      % return a file descriptor used
%      % in later references to the
%      % file. Channel used only
%                                             % if needed to generate FileDesc
%  SysWriteRec (FileDescriptor, StringToWrite, StringLength);
%      % Write StringLength characters
%      % from StringToWrite from the
%      % first position.
%  LengthRead := SysReadRec (FileDescriptor, StringBuffer);
%      % Read from the FileDescriptor, a
%      %  record into the StringBuffer.
%      %  Return the length of the
%      %  string read.
%  SysClose (FileDescriptor);      % Close FileDescriptor, allowing
%      %  it to be reused.
%
%==============================================================================
 
(compiletime
  (load fast-vector if-system io-decls))
 
(on fast-integers)
 
(fluid '(in* out*))
 
(fluid '(stdin* bregsave
       *parens parencount*
       stdout*
       errout*
       promptout*
       *echo
       *staticstring*
       iobuffer ))
 
(de findfreechannel ()
 
  %  Look into the ChannelStatus array for a free channel.
 
  (let ((channel 0))
    (while (neq (wgetv channelstatus channel) channelclosed)
      (when (wgreaterp channel maxchannels)
      (ioerror "No free channels left"))
      (setq channel (wplus2 channel 1)))
    channel
    ))
 
 
(de systemopenfileforinput (filename)
 
  %   Open the argument filename as a read only file.
 
  (let ((channel (findfreechannel)))
    (setf (wgetv channeltable  channel) (sysopenread channel filename))
    (setf (wgetv channelstatus channel) channelopenread)
    (setf (wgetv maxbuffer     channel) (sysmaxbuffer (wgetv channeltable channel)))
    (setf (wgetv readfunction  channel) 'independentreadchar)
    (setf (wgetv writefunction channel) 'readonlychannel)
    (setf (wgetv closefunction channel) 'independentclosechannel)
    (setf (wgetv nextposition  channel) 0)   % Will be post Incremented
    (setf (wgetv bufferlength  channel) -1)
    (setf (words-fetch iobuffer channel)(mkstring (wgetv maxbuffer channel) 32))
    channel
    ))
 
 
(de systemopenfileforoutput (filename)
 
  % Open the argument file name as a write only file.
 
  (let ((channel (findfreechannel)))
    (setf (wgetv channeltable  channel) (sysopenwrite channel filename))
    (setf (wgetv channelstatus channel) channelopenwrite)
    (setf (wgetv maxbuffer     channel) (sysmaxbuffer (wgetv channeltable channel)))
    (setf (wgetv readfunction  channel) 'writeonlychannel)
    (setf (wgetv writefunction channel) 'independentwritechar)
    (setf (wgetv closefunction channel) 'independentclosechannel)
    (setf (words-fetch iobuffer channel) (mkstring (wgetv maxbuffer channel) 32))
    (setf (wgetv nextposition  channel) -1)   % Will be set pre-incremented
    (setf (wgetv bufferlength  channel) (wgetv maxbuffer channel))
    channel
    ))
 
 
(de systemopenfilespecial (filename)
 
  %  Mark a channel as open for a special purpose.
 
  (let ((channel (findfreechannel)))
    (setf (wgetv channelstatus channel) channelopenspecial)
    channel
    ))
 
 
(de testlegalchannel (channel)
 
  % Cause an error if CHANNEL is not legal.
 
  (unless (and (posintp channel) (wlessp channel maxchannels))
    (ioerror (list channel " is not a legal channel "))))
 
(compiletime
(ds opentestlegalchannel (channel)
 
  % Cause an error if CHANNEL is not legal.
 
  (unless (and (posintp channel) (wlessp channel maxchannels))
             (ioerror (list channel " is not a legal channel "))))
)
 
(compiletime
(flag '(inputchannelerror) 'internalfunction))
 
(compiletime
(ds opentestlegalchannelandreturniferror (channel)
 
  % Cause an error if CHANNEL is not legal.
 
  (unless (eq channel (wand channel maxchannels)) % must be 2**n -1
    (return (inputchannelerror channel)))))
 
(compiletime (flag '(independentreadchar-hardcase bldmsg1)
                                                   'internalfunction))
 
(de independentreadchar (channel)
 
  % This function will read in a character from the buffer.  It will read
  % the record on buffer length overflow only.  Thus when an EOL character
  % is read, it is processed as any other character, except, if it is the last
  % one, in the record, it will do the read automatically.
  % Note, this will not read the next record until after the final character
  % has been processed.
 
 (prog (nextpo chr bufl echo)
  (opentestlegalchannelandreturniferror  channel)
  (setq nextpo (wgetv nextposition channel))
  (setq bufl   (wgetv bufferlength channel))
  (setq echo *echo)
 
  (when (or (wgreaterp nextpo bufl) echo)
      (return (independentreadchar-hardcase channel nextpo bufl echo)))
 
  (setq chr (strbyt (words-fetch iobuffer channel) nextpo))
 
  (setf (wgetv nextposition channel) (wplus2 nextpo 1))
  (return chr)
))
 
(de independentreadchar-hardcase (channel nextpo bufl echo)
 
  % If the input buffer is empty, read another buffer full:
 
  (when (wgreaterp nextpo bufl)
 
    (when (and *parens echo)
          (channelwritestring out*
                 (bldmsg1 (if (wlessp parencount* 10)
                              " /%w/ " "/%w/ ") parencount*)))
    (setf (wgetv bufferlength channel)
          (sysreadrec (wgetv channeltable channel)
                                   (words-fetch iobuffer channel)))
    (setq nextpo 0)
    (setf (wgetv nextposition channel) 0))
 
  % Pull the next character out of the buffer.
 
  (let ((chr (strbyt (words-fetch iobuffer channel) nextpo)))
 
    (setf (wgetv nextposition channel) (wplus2 nextpo 1))
    (when echo    %dont produce control chars!
     (Writechar (if (and (wneq chr 10) (wneq chr 9) %or lf or tab
			 (wlessp chr 32)) 35 chr)))  % #
    chr
  ))
 
(if_system Cray1
 (de independentwritechar (channel chr)
 
  % Write a character into the buffer.  Actually dump the buffer when the
  % EOL character is found, or when the buffer is full.  This happens
  % immediately upon meeting this condition, not waiting for the
  % next character.  Note, that this places the EOL character into the
  % buffer for machine dependent treatment as CR/LF etc
  % preloading for CRAY !!!
 
 (prog (nextpo iobuff buffl x)
  (opentestlegalchannelandreturniferror  channel)
  (setq nextpo (wgetv nextposition channel))
  (setq iobuff (words-fetch iobuffer channel))
  (setq buffl  (wgetv bufferlength channel))      % Number of characters
 
  (setq nextpo (wplus2 nextpo 1))
  (setq x (wrshift (wor (wdifference chr (wplus2 (char eol) 1))
                        (wdifference nextpo buffl))
               63))
  (setf (wgetv nextposition channel) nextpo )
  (setf (strbyt iobuff nextpo) chr)
  (setq nextpo (iadd1 nextpo))
  (cond ((eq 0 x) (return nil))
        ((or (weq chr (char eol))
            (wgeq nextpo buffl))  % Indexed from 0 , exact criterion
               (return (flushbuffer channel)))
        (t (return 0)))
))
(de independentwritechar (channel chr)
 
  % Write a character into the buffer.  Actually dump the buffer when the
  % EOL character is found, or when the buffer is full.  This happens
  % immediately upon meeting this condition, not waiting for the
  % next character.  Note, that this places the EOL character into the
  % buffer for machine dependent treatment as CR/LF etc
  % preloading for CRAY !!!
 
 (prog (nextpo iobuff buffl x)
  (DISable-operand-range-interrupt)
  (setq nextpo (wgetv nextposition channel))
  (setq iobuff (words-fetch iobuffer channel))
  (setq buffl  (wgetv bufferlength channel))      % Number of characters
  (opentestlegalchannelandreturniferror  channel)
  (enable-operand-range-interrupt)
 
  (setq nextpo (wplus2 nextpo 1))
  (setq x (wrshift (wor (wdifference chr (wplus2 (char eol) 1))
                        (wdifference nextpo buffl))
               63))
  (setf (wgetv nextposition channel) nextpo )
  (setf (strbyt iobuff nextpo) chr)
  (setq nextpo (iadd1 nextpo))
  (cond ((eq 0 x) (return nil))
        ((or (weq chr (char eol))
            (wgeq nextpo buffl))  % Indexed from 0 , exact criterion
               (return (flushbuffer channel)))
        (t (return 0)))
))
)
 
(de flushbuffer (channel)
 
  % Actually output any characters still hanging around in the buffer.
 
  (syswriterec (wgetv channeltable channel)
             (words-fetch iobuffer channel)
             (wgetv nextposition channel))
  (setf (wgetv nextposition channel) -1)
  )
 
 
(de systemmarkasclosedchannel (channel)
 
  % Mark the argument channel as closed and update the read, write and
  % close functions likewise.  Careful, if the caller does this first
  % and then trys to access a read, write or close function we are
  % in big trouble.  Is it correct to do this?????  Or is a marking of
  % the channel status table sufficient.
 
  (testlegalchannel channel)
  (setf (wgetv channelstatus channel) channelclosed)
  (setf (wgetv readfunction  channel) 'channelnotopen)
  (setf (wgetv writefunction channel) 'channelnotopen)
  (setf (wgetv closefunction channel) 'channelnotopen)
  )
 
 
(de independentclosechannel (channel)
 
  % Actually close the argument channel.
 
  (testlegalchannel channel)
  (sysclose (wgetv channeltable channel))
  )
 
 
(de clearonechannel (chn filedes how fn)
  % Initialize Channel Tables, etc for initial "hand-built" channels
  % Assume these not CLOSEABLE
 
  (let ((bufflen  (sysmaxbuffer filedes)))
 
    (setf (wgetv channeltable  chn) filedes)
    (setf (wgetv maxbuffer     chn) bufflen)
    (setf (wgetv nextposition  chn) 0)
    (setf (wgetv lineposition  chn) 0)
    (setf (wgetv pageposition  chn) 0)
    (setf (wgetv unreadbuffer  chn) 0)
    (setf (wgetv closefunction chn)  'illegalstandardclosechannel)
 
    (if (weq how 'input)
      (progn (setf (wgetv bufferlength   chn) -1)
           (setf (wgetv channelstatus  chn)  channelopenread)
           (setf (wgetv readfunction   chn) fn)
           (setf (wgetv writefunction  chn) 'readonlychannel)
           )
      (progn (setf (wgetv nextposition   chn) -1)
           (setf (wgetv bufferlength   chn) (wgetv maxbuffer chn))
           (setf (wgetv channelstatus  chn)  channelopenwrite)
           (setf (wgetv readfunction   chn) 'writeonlychannel)
           (setf (wgetv writefunction  chn) fn)
           )
      )
    (setf (words-fetch iobuffer chn) (mkstring bufflen 32))
    ))
 
 
(de cleario nil
  (unless (vectorp iobuffer)
    (setf iobuffer (mkvect (wplus2 maxchannels 1))) %
    )                                              %
  (syscleario)                                     %
  (setf in*  stdin*)
  (setf out* stdout*)
  )
 
 
%     2/24/84 - rrk We are getting an extra carriage return before printing
% the result of every evaluation of the top loop.  This is caused because
% of our method of handling prompts.  We write a prompt to the the
% promptout channel.  Write is the only thing that updates the lineposition
% numbers.  Thus, when we come around to printout the results, it looks
% like the prompt has reset the lineposition (length prompstring).  In some
% implementations, promptout and terminal out are the same channel, thus
% the call to posn returns something non zero.  This will not happen when
% the channel numbers are different.  Therefore to make it all work
% together, the following line will check to see if the two channels are
% the same, and reset the lineposition to 0 if they are.  We could have
% ignored the test in favor of always setting the position to 0, but that
% may be inconsistent with the philosophy that all writes update the
% lineposition (however one could argue that the prompt write is special
% and should not update the lineposition).  I'll leave the test in, and
% leave it to future bug reports if this is not what is expected.
 
(compiletime (flag '(terminalinputhandler-hardcase bldmsg1)
                                                    'internalfunction))
(If_system Cray1
(de terminalinputhandler (channel)
 (prog (nextpo chr bufl echo iobuf)
  (opentestlegalchannelandreturniferror  channel)
  (setq nextpo (wgetv nextposition channel))
  (setq bufl   (wgetv bufferlength channel))
  (setq echo *echo)
  (setq iobuf (words-fetch iobuffer channel))
  (when (or (wgreaterp nextpo bufl) echo)
   (return (terminalinputhandler-hardcase channel nextpo bufl echo iobuf
            )))
  (setq  chr  (strbyt iobuf nextpo ))
  (setf (wgetv nextposition channel)  (wplus2 nextpo 1))
  (return chr)
))
(de terminalinputhandler (channel)
 (prog (nextpo chr bufl echo iobuf)
  (DISable-operand-range-interrupt)
  (setq nextpo (wgetv nextposition channel))
  (setq bufl   (wgetv bufferlength channel))
  (setq echo *echo)
  (setq iobuf (words-fetch iobuffer channel))
  (opentestlegalchannelandreturniferror  channel)
  (enable-operand-range-interrupt)
  (when (or (wgreaterp nextpo bufl) echo)
   (return (terminalinputhandler-hardcase channel nextpo bufl echo iobuf
            )))
  (setq  chr  (strbyt iobuf nextpo ))
  (setf (wgetv nextposition channel)  (wplus2 nextpo 1))
  (return chr)
)))
 
(de inputchannelerror (channel)
   (prog (z)
    (setq z (list channel
                  " is not a legal channel" ))
    (enable-operand-range-interrupt)
    (return (ioerror z))))
 
 
(de terminalinputhandler-hardcase (channel nextpo buflen echo iobuf)
 
  (when (wgreaterp nextpo buflen)
   (when (or echo (not *batch))
    (channelwritestring promptout*
           (if (stringp promptstring*)
               (if  *parens
                    (if (wlessp parencount* 10)
                         (bldmsg1 " /%w/ %w" parencount* promptstring*)
                         (bldmsg1 "/%w/ %w" parencount* promptstring*))
                    promptstring*)
                             ">")))
    (when (not *batch) (flushbuffer promptout*))
    (when (weq promptout* stdout*)
      (setf (wgetv lineposition promptout*) 0))
    (setf (wgetv bufferlength channel)
      (sysreadrec (wgetv channeltable channel) iobuf))
    (setq nextpo 0)
    (setf (wgetv nextposition channel) 0))
 
  (let ((chr  (strbyt iobuf nextpo )))
    (setf (wgetv nextposition channel)  (wplus2 nextpo 1))
    (when echo    %dont produce control chars!
     (channelwritechar out*
                (if (and (wneq chr 10)  (wneq chr 9) %or lf or tab
			 (wlessp chr 32)) 35 chr)))  % #
      chr
  ) )
 
(setq *staticstring* (gtconststr 50))
 
(De Bldmsg1 (Format Ar1 Ar2 Ar3 Ar4)
 
 % problem: printing parenthesis depth counter
 % cant use tokenbuffer, because this will destroy things that cross
 % over line boundary, use BREGSAVE instead, it is huge but is will
 % be destroyed by garbage collector. What a mess!
 
  (Prog (tempchannel Temporarytokenbuffer)
        (setf (Wgetv Lineposition 2) 0)
        (setq tempchannel out*)
        (setq Temporarytokenbuffer tokenbuffer)
        (setq tokenbuffer BregSave)
        (setf (wgetv tokenbuffer 0) -1)
        (setf out* 2)
        (printf Format Ar1 Ar2 Ar3 Ar4)
        (setf (Strbyt tokenbuffer
                      (Wplus2 (Wgetv tokenbuffer 0) 1))
              (Char null))
        (setf  out* tempchannel)
        (setq tokenbuffer Temporarytokenbuffer)
        (setf (wgetv *staticstring* 0) (wgetv bregsave 0)) %shorten
        (Return (Copystringtofrom *staticstring* Bregsave))))
 
(off fast-integers)
